-------------------------------------------------------------------------------
-- Localized Lua globals.
-------------------------------------------------------------------------------
local _G = getfenv(0)

-- Functions

-- Libraries
local string = _G.string
local table = _G.table


-------------------------------------------------------------------------------
-- AddOn namespace.
-------------------------------------------------------------------------------
local FOLDER_NAME, private = ...

local LibStub = _G.LibStub
local addon = LibStub("AceAddon-3.0"):NewAddon(FOLDER_NAME, "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")

private.pet_list = {}
private.owned_list = {}
private.unowned_list = {}

local known_pets_by_name = {}
local known_pets_by_id = {}
local unknown_pets_by_name = {}
local unknown_pets_by_id = {}


-------------------------------------------------------------------------------
-- Constants.
-------------------------------------------------------------------------------
local MAX_ACTIVE_PETS = 3


-------------------------------------------------------------------------------
-- Helper functions.
-------------------------------------------------------------------------------
local AcquireTable
local ReleaseTable
do
	local table_cache = {}


	function AcquireTable()
		return table.remove(table_cache) or {}
	end


	function ReleaseTable(tbl)
		if not tbl then
			return
		end
		table.wipe(tbl)
		table.insert(table_cache, tbl)
	end
end -- do block


local ModifyTooltip
do
	local output = {}
	local temp = {}

	function ModifyTooltip(identifier, known_table, unknown_table)
		local known_entries = known_table[identifier]

		table.wipe(output)

		if known_entries then
			local shown_header

			for index = 1, #known_entries do
				local pet = known_entries[index]
				table.wipe(temp)

				if not shown_header then
					table.insert(temp, "\n")
					table.insert(temp, pet.name)
					table.insert(temp, _G.HEADER_COLON)

					if #known_entries < MAX_ACTIVE_PETS then
						table.insert(temp, "\n")
						table.insert(temp, "|cff82c5ff")
						table.insert(temp, _G.UNIT_CAPTURABLE)
						table.insert(temp, "|r")
					end
					table.insert(temp, "\n|cffffffff")
					table.insert(temp, _G.BATTLE_PET_CAPTURED)
					table.insert(temp, _G.HEADER_COLON)
					table.insert(temp, "|r\n")
					shown_header = true
				end
				table.insert(temp, _G.ITEM_QUALITY_COLORS[pet.rarity - 1].hex)
				table.insert(temp, "    ")
				table.insert(temp, _G["BATTLE_PET_BREED_QUALITY" .. pet.rarity])
				table.insert(temp, "|r |cffffffff")
				table.insert(temp, _G.PARENS_TEMPLATE:format(_G.TOOLTIP_UNIT_LEVEL:format(pet.level)))
				table.insert(temp, "|r")

				table.insert(output, table.concat(temp, ""))
			end
		elseif unknown_table[identifier] then
			local pet = unknown_table[identifier][1]
			table.wipe(temp)

			table.insert(temp, "\n")
			table.insert(temp, pet.name)
			table.insert(temp, _G.HEADER_COLON)
			table.insert(temp, "\n")
			table.insert(temp, "|cff82c5ff")
			table.insert(temp, _G.UNIT_CAPTURABLE)
			table.insert(temp, "|r")

			table.insert(output, table.concat(temp, ""))
		end

		if #output > 0 then
			_G.GameTooltip:AddLine(table.concat(output, "\n"))
		end
	end
end

function private.UpdatePetList()
	for index = 1, #private.pet_list do
		ReleaseTable(private.pet_list[index])
	end
	table.wipe(private.pet_list)
	table.wipe(private.owned_list)
	table.wipe(private.unowned_list)
	table.wipe(known_pets_by_id)
	table.wipe(known_pets_by_name)

	local num_pets = _G.C_PetJournal.GetNumPets(_G.PetJournal.isWild)

	for pet_index = 1, num_pets do
		local pet_id, _, is_owned, _, level, _, _, name, _, pet_type, creature_id, _, _, _, can_battle = _G.C_PetJournal.GetPetInfoByIndex(pet_index, _G.PetJournal.isWild)
		local _, max_health, attack, speed, rarity = _G.C_PetJournal.GetPetStats(pet_id)

		local pet = AcquireTable()
		pet.attack = attack
		pet.can_battle = can_battle
		pet.id = pet_id
		pet.is_owned = is_owned
		pet.journal_index = pet_index
		pet.level = level
		pet.max_health = max_health
		pet.name = name
		pet.rarity = rarity
		pet.speed = speed
		pet.type = pet_type

		private.pet_list[#private.pet_list + 1] = pet

		if is_owned then
			known_pets_by_id[creature_id] = known_pets_by_id[creature_id] or {}
			known_pets_by_id[creature_id][#known_pets_by_id[creature_id] + 1] = pet

			known_pets_by_name[name] = known_pets_by_name[name] or {}
			known_pets_by_name[name][#known_pets_by_name[name] + 1] = pet

			private.owned_list[#private.owned_list + 1] = pet
		else
			unknown_pets_by_id[creature_id] = unknown_pets_by_id[creature_id] or {}
			unknown_pets_by_id[creature_id][#unknown_pets_by_id[creature_id] + 1] = pet

			unknown_pets_by_name[name] = unknown_pets_by_name[name] or {}
			unknown_pets_by_name[name][#unknown_pets_by_name[name] + 1] = pet

			private.unowned_list[#private.unowned_list + 1] = pet
		end
	end
	-- TODO: Sort here.
--	private.PetJournal_FindPetCardIndex()
--	private.PetJournal_UpdatePetList()
	_G.PetJournal_FindPetCardIndex()
	_G.PetJournal_UpdatePetList()
end


-------------------------------------------------------------------------------
-- Initialization functions.
-------------------------------------------------------------------------------
function addon:OnInitialize()
	self:RegisterEvent("ADDON_LOADED")
end


function addon:OnEnable(event_name)
	do
		local previous_tooltip_line
		local entity_registry = {}

		_G.GameTooltip:HookScript("OnUpdate", function(self)
			if not _G.Minimap:IsMouseOver() then
				return
			end
			local tooltip_line = _G["GameTooltipTextLeft1"]:GetText()

			if not tooltip_line or tooltip_line == previous_tooltip_line then
				return
			end
			previous_tooltip_line = tooltip_line
			tooltip_line = tooltip_line:gsub([[|T(.-)|t]], "")

			if tooltip_line:find("\n") then
				local entities = { ("\n"):split(tooltip_line) }
				table.wipe(entity_registry)

				for index = 1, #entities do
					local entity = entities[index]

					if not entity_registry[entity] then
						ModifyTooltip(entity, known_pets_by_name, unknown_pets_by_name)
						entity_registry[entity] = true
					end
				end
			else
				ModifyTooltip(tooltip_line, known_pets_by_name, unknown_pets_by_name)
			end
			_G.GameTooltip:Show()
		end)

		_G.GameTooltip:HookScript("OnHide", function()
			previous_tooltip_line = nil
		end)
	end
	-- Trickery to get the the initial pet list.
	self:ScheduleTimer(function()
		_G.TogglePetJournal(2)
		_G.HideUIPanel(_G.PetJournalParent)
	end, 2)
end


-------------------------------------------------------------------------------
-- Event handlers.
-------------------------------------------------------------------------------
function addon:ADDON_LOADED(event_name, addon_name)
	if addon_name ~= "Blizzard_PetJournal" then
		return
	end
	self:HookScript(_G.PetJournalParent, "OnShow", private.UpdatePetList)
--	self:RawHook("PetJournal_FindPetCardIndex", private.PetJournal_FindPetCardIndex, true)
--	self:RawHook("PetJournal_UpdatePetList", private.PetJournal_UpdatePetList, true)

	self:UnregisterEvent("ADDON_LOADED")
	self:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
end


function addon:PET_JOURNAL_LIST_UPDATE(event_name)
	private.UpdatePetList()
end


function addon:UPDATE_MOUSEOVER_UNIT()
	if not _G.UnitIsWildBattlePet("mouseover") then
		return
	end
	ModifyTooltip(_G.tonumber(_G.UnitGUID("mouseover"):sub(-12, -9), 16), known_pets_by_id, unknown_pets_by_id)
end
